from django.contrib.admin.options import get_content_type_for_model
from django.db import models

from ..filters import FilterMixin
from .abc import Resolver
from .mixins import CreateModelMixin, DestroyModelMixin, ListModelMixin, UpdateModelMixin


class GenericModelResolver(Resolver):
    """
    Model based resolver based off of DRF model views.
    """

    queryset = None
    model = None
    nested_field_name = None

    def get_queryset(self):
        assert self.queryset is not None, (
            "'%s' should either include a `queryset` attribute, "
            "or override the `get_queryset()` method." % self.__class__.__name__
        )

        queryset = self.queryset
        if isinstance(queryset, models.query.QuerySet):
            # Ensure queryset is re-evaluated on each request.
            queryset = queryset.all()
        if self.config.get("nested", False):
            queryset = self.filter_nested_queryset(queryset).all()
        return queryset

    def get_model(self):
        if self.model is not None:
            return self.model
        else:
            return self.get_queryset().model

    def filter_nested_queryset(self, queryset):
        # see if we can just use the info context and the name of the resolver field
        # to reverse the relation, fallback on parent_name
        try:
            return getattr(
                self.parent,
                self.config.get("nested_field_name", self.nested_field_name)
                or self.info.field_name,
            )
        except AttributeError:
            params = {}
            if self.config.get("parent_name", None) == "object_id":
                params["object_id"] = self.parent.id
                params["content_type"] = get_content_type_for_model(self.parent._meta.model)
            elif self.config.get("parent_name", None):
                params[self.config.get("parent_name")] = self.parent.id

            return queryset.filter(**params)


class ModelResolver(
    CreateModelMixin, UpdateModelMixin, DestroyModelMixin, FilterMixin, GenericModelResolver
):
    pass


class ListModelResolver(ListModelMixin, GenericModelResolver):
    pass



from functools import update_wrapper

import humps.main as humps
from django.utils.decorators import classonlymethod

from ariadne_extended.settings import api_settings

from . import exceptions


class Resolver:
    permission_classes = api_settings.DEFAULT_PERMISSION_CLASSES
    throttle_classes = []
    authentication_classes = []
    default_method = "retrieve"

    def __init__(self, parent, info, *operation_args, config=dict(), **operation_kwargs):
        # config is used for this specific operation on the resolver
        self.config = config.copy()
        # Graphql info object stored for access
        self.info = info
        # Get request from graphql info by default
        self.request = self.get_request(info)
        # Store parent for access
        self.parent = parent

        # Store unprocessed resolver args and kwargs
        self._operation_args = operation_args
        self._operation_kwargs = operation_kwargs

        # process resolver args and kwargs if needed
        self.operation_args = self.get_operation_args()
        self.operation_kwargs = self.get_operation_kwargs()

        # normalize federation reference args from ariadne
        self.reference_kwargs = self.get_reference_kwargs()

    def get_request(self, info):
        context = getattr(info, "context", {})
        return context.get("request", None)

    def initial(self, request, *args, **kwargs):
        """
        Runs anything that needs to occur prior to calling the operation handler
        """
        self.perform_authentication(request)
        self.check_permissions(request)
        self.check_throttles(request)

    def resolve(self, parent, *args, **kwargs):
        """
        Run initial checks, then call the default or configured operation resolve handler
        """
        self.initial(self.request, *args, **kwargs)

        method = self.config.get("method", self.default_method)
        handler = getattr(self, method)
        try:
            return handler(parent, *args, **kwargs)
        except exceptions.ResolverException:
            # TODO: tac on graphql errors?
            return None

    def get_operation_args(self):
        return self._operation_args

    def get_operation_kwargs(self):
        op_values = self._operation_kwargs.copy()

        # convert camelcase to snake case on all
        return humps.decamelize(op_values)

    def get_reference_kwargs(self):
        if self.config.get("reference", False):
            return humps.decamelize(self.parent)
        else:
            return dict()

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs["context"] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.

        You may want to override this if you need to provide different
        serializations depending on the incoming request.

        (Eg. admins get full serialization, others get basic serialization)
        """
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method." % self.__class__.__name__
        )

        return self.serializer_class

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {"request": self.request, "format": None, "view": self}

    @classonlymethod
    def as_resolver(cls, **resolver_config):
        def resolver(parent, info, *args, **kwargs):
            self = cls(parent, info, config=resolver_config, *args, **kwargs)
            return self.resolve(parent, *self.operation_args, **self.operation_kwargs)

        resolver.resolver_class = cls

        # take name and docstring from class
        update_wrapper(resolver, cls, updated=())

        # and possible attributes set by decorators
        update_wrapper(resolver, cls.resolve, assigned=())

        return resolver

    @classonlymethod
    def as_nested_resolver(cls, **resolver_config):
        """
        When resolving a related type within another type. Performs needed config changes
        to trigger nested filtering between related objects and select the correct relational field
        from the orm when used inside a GenericModelResolver
        """
        resolver_config["nested"] = True
        return cls.as_resolver(**resolver_config)

    @classonlymethod
    def as_reference_resolver(cls, **resolver_config):
        """
        When resolving via reference this enables the resolver to grab the correct reference
        parameters from ariadne and utilize them in subsequent object lookups.
        """
        resolver_config["reference"] = True
        return cls.as_resolver(**resolver_config)

    def permission_denied(self, request, message=None):
        """
        If resolution is not permitted, determine what kind of exception to raise.
        """
        raise exceptions.PermissionDenied(message)

    def check_object_permissions(self, request, obj):
        """
        Check if the resolution should be permitted for a given object.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self.get_permissions():
            if not permission.has_object_permission(request, self, obj):
                self.permission_denied(request, message=getattr(permission, "message", None))

    def check_throttles(self, request):
        """
        Check if resolution should be throttled.
        Raises an appropriate exception if the resolution is throttled.
        """
        throttle_durations = [
            throttle.wait()
            for throttle in self.get_throttles()
            if not throttle.allow_request(request, self)
        ]

        if throttle_durations:
            # Filter out `None` values which may happen in case of config / rate
            # changes, see #1438
            durations = [duration for duration in throttle_durations if duration is not None]

            duration = max(durations, default=None)
            self.throttled(self.request, duration)

    def perform_authentication(self, request):
        """
        Perform authentication on the incoming resolution.

        Note that if you override this and simply 'pass', then authentication
        will instead be performed lazily, the first time either
        `request.user` or `request.auth` is accessed.
        """
        pass

    def check_permissions(self, request):
        """
        Check if the resolution should be permitted.
        Raises an appropriate exception if the resolution is not permitted.
        """
        for permission in self.get_permissions():
            if not permission.has_permission(request, self):
                self.permission_denied(request, message=getattr(permission, "message", None))

    def get_authenticators(self):
        """
        Instantiates and returns the list of authenticators that this resolver can use.
        """
        return [auth() for auth in self.authentication_classes]

    def get_permissions(self):
        """
        Instantiates and returns the list of permissions that this resolver requires.
        """
        return [permission() for permission in self.permission_classes]

    def get_throttles(self):
        """
        Instantiates and returns the list of throttles that this resolver uses.
        """
        return [throttle() for throttle in self.throttle_classes]

    def throttled(self, request, wait):
        """
        If request is throttled, determine what kind of exception to raise.
        """
        raise exceptions.Throttled(wait)



def patch_autoreload():
    """
    Monkey patches Django's autoreload file finder to pickup and trigger when graphql files
    are added or changed.

    Only for development use, intended to be called before running main() in your manage.py.
    """
    try:
        import django.utils.autoreload
    except Exception:
        return
    from glob import glob
    from itertools import chain
    from pathlib import Path

    old_iter = django.utils.autoreload.iter_all_python_module_files

    def new_iter():
        results = old_iter()
        graphql_files = map(Path, glob("**/**/*.graphql"))
        return frozenset(chain(results, graphql_files))

    django.utils.autoreload.iter_all_python_module_files = new_iter



import json
from typing import Optional, cast

from ariadne.contrib.django.views import GraphQLView
from ariadne.contrib.tracing.apollotracing import ApolloTracingExtensionSync
from ariadne.format_error import format_error
from ariadne.graphql import graphql_sync
from ariadne.types import ContextValue, ErrorFormatter, GraphQLResult, RootValue
from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest, HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from graphql import GraphQLSchema
from graphql import format_error as format_graphql_error
from graphql.error import GraphQLError
from graphql.execution import MiddlewareManager


@method_decorator(csrf_exempt, name="dispatch")
class BaseGraphQLView(GraphQLView):
    extensions = [ApolloTracingExtensionSync]

    def execute_query(self, request: HttpRequest, data: dict) -> GraphQLResult:
        """TODO: remove when extensions is added to graphql_sync in parent view"""
        if callable(self.context_value):
            context_value = self.context_value(request)  # pylint: disable=not-callable
        else:
            context_value = self.context_value or request

        return graphql_sync(
            cast(GraphQLSchema, self.schema),
            data,
            context_value=context_value,
            root_value=self.root_value,
            validation_rules=self.validation_rules,
            debug=settings.DEBUG,
            logger=self.logger,
            error_formatter=self.error_formatter or format_error,
            extensions=self.extensions,
            middleware=self.middleware,
        )


@method_decorator(csrf_exempt, name="dispatch")
class LoginRequiredGraphQLView(LoginRequiredMixin, BaseGraphQLView):
    def handle_no_permission(self):
        return HttpResponseForbidden()


class HttpError(Exception):
    def __init__(self, response, message=None, *args, **kwargs):
        self.response = response
        self.message = message = message or response.content.decode()
        super(HttpError, self).__init__(message, *args, **kwargs)


@method_decorator(csrf_exempt, name="dispatch")
class BatchGraphQLView(LoginRequiredMixin, View):
    """
    Batch GraphQL view based on the graphene batch view.
    """

    # FIX: GraphQL errors aren't bubbling up properly, probably due to the differences with Graphene
    http_method_names = ["get", "post", "options"]
    schema: Optional[GraphQLSchema] = None
    context_value: Optional[ContextValue] = None
    root_value: Optional[RootValue] = None
    logger = None
    validation_rules = None
    error_formatter: Optional[ErrorFormatter] = None
    middleware: Optional[MiddlewareManager] = None
    batch = False
    pretty = False

    def handle_no_permission(self):
        return HttpResponseForbidden()

    def get(self, request: HttpRequest, *args, **kwargs):
        return self.handle_query(request, *args, **kwargs)

    def post(self, request: HttpRequest, *args, **kwargs):
        return self.handle_query(request, *args, **kwargs)

    def handle_query(self, request, *args, **kwargs):
        if not self.schema:
            raise ValueError("GraphQLView was initialized without schema.")
        try:
            data = self.extract_data_from_request(request)

            if self.batch:
                responses = [self.get_response(request, entry) for entry in data]
                result = "[{}]".format(",".join([response[0] for response in responses]))
                status_code = (
                    responses and max(responses, key=lambda response: response[1])[1] or 200
                )
            else:
                result, status_code = self.get_response(request, data)

            return HttpResponse(
                status=status_code, content=result, content_type="application/json"
            )

        except HttpError as e:
            # TODO: return actual JsonResponse
            response = e.response
            response["Content-Type"] = "application/json"
            response.content = self.json_encode(request, {"errors": [self.format_error(e)]})
            return response

    def get_response(self, request, data):
        params = self.get_graphql_params(request, data)

        execution_result_passed, execution_result = self.execute_graphql_request(request, params)

        status_code = 200
        response = {}
        if execution_result_passed:
            if execution_result.get("errors", False):
                status_code = 400
            else:
                response["data"] = execution_result["data"]

        else:
            status_code = 400
            response["errors"] = [self.format_error(e) for e in execution_result["errors"]]

        if self.batch:
            response["id"] = params["id"]
            response["status"] = status_code

        result = self.json_encode(request, response)
        return result, status_code

    def extract_data_from_request(self, request):
        content_type = self.get_content_type(request)

        if content_type == "application/graphql":
            return {"query": request.body.decode()}

        elif content_type == "application/json":
            # noinspection PyBroadException
            try:
                body = request.body.decode("utf-8")
            except Exception as e:
                raise HttpError(HttpResponseBadRequest(str(e)))

            try:
                request_json = json.loads(body)
                if self.batch:
                    assert isinstance(request_json, list), (
                        "Batch requests should receive a list, but received {}."
                    ).format(repr(request_json))
                    assert len(request_json) > 0, "Received an empty list in the batch request."
                else:
                    assert isinstance(
                        request_json, dict
                    ), "The received data is not a valid JSON query."
                return request_json
            except AssertionError as e:
                raise HttpError(HttpResponseBadRequest(str(e)))
            except (TypeError, ValueError):
                raise HttpError(HttpResponseBadRequest("POST body sent invalid JSON."))

        elif content_type in ["application/x-www-form-urlencoded", "multipart/form-data"]:
            return request.POST

        return {}

    def execute_graphql_request(self, request, data):
        if callable(self.context_value):
            context_value = self.context_value(request)
        else:
            context_value = self.context_value or request

        return graphql_sync(
            cast(GraphQLSchema, self.schema),
            data,
            context_value=context_value,
            root_value=self.root_value,
            debug=settings.DEBUG,
            logger=self.logger,
            validation_rules=self.validation_rules,
            error_formatter=self.error_formatter or format_error,
            middleware=self.middleware,
        )

    @staticmethod
    def get_graphql_params(request, data):
        query = request.GET.get("query") or data.get("query")
        variables = request.GET.get("variables") or data.get("variables")
        id = request.GET.get("id") or data.get("id")

        if variables and isinstance(variables, str):
            try:
                variables = json.loads(variables)
            except Exception:
                raise HttpError(HttpResponseBadRequest("Variables are invalid JSON."))

        operation_name = request.GET.get("operationName") or data.get("operationName")
        if operation_name == "null":
            operation_name = None

        return dict(query=query, variables=variables, operation_name=operation_name, id=id)

    @staticmethod
    def format_error(error):
        if isinstance(error, GraphQLError):
            return format_graphql_error(error)

        return {"message": str(error)}

    @staticmethod
    def get_content_type(request):
        meta = request.META
        content_type = meta.get("CONTENT_TYPE", meta.get("HTTP_CONTENT_TYPE", ""))
        return content_type.split(";", 1)[0].lower()

    def json_encode(self, request, data, pretty=False):
        if not (self.pretty or pretty) and not request.GET.get("pretty"):
            return json.dumps(data, separators=(",", ":"))

        return json.dumps(data, sort_keys=True, indent=2, separators=(",", ": "))



from .pagination import CursorPaginator
from ..resolvers.mixins import ListModelMixin


class RelayModelMixin(ListModelMixin):
    pagination_class = CursorPaginator

    def get_ordering(self):
        if hasattr(self, "ordering"):
            return self.ordering
        return ("id",)

    def paginate_queryset(self, queryset, **kwargs):
        kwargs.update(
            dict(
                ordering=self.get_ordering(),
                first=self.operation_kwargs.get("first", None),
                last=self.operation_kwargs.get("last", None),
                after=self.operation_kwargs.get("after", None),
                before=self.operation_kwargs.get("before", None),
            )
        )
        return super().paginate_queryset(queryset, **kwargs)

    def list(self, request, *args, **kwargs):
        page = super().list(request, *args, **kwargs)
        return {
            "edges": [{"cursor": page.cursor(node), "node": node} for node in page],
            "pageInfo": page,
        }



from django_filters.rest_framework import DjangoFilterBackend


class GraphQLFilterBackend(DjangoFilterBackend):
    """Provide filter data from the resolver to the filterset"""

    def get_filterset_kwargs(self, request, queryset, resolver):
        return {"data": resolver.get_filter_data(), "queryset": queryset, "request": request}



"""
Provides a set of pluggable permission policies.
"""


class OperationHolderMixin:
    def __and__(self, other):
        return OperandHolder(AND, self, other)

    def __or__(self, other):
        return OperandHolder(OR, self, other)

    def __rand__(self, other):
        return OperandHolder(AND, other, self)

    def __ror__(self, other):
        return OperandHolder(OR, other, self)

    def __invert__(self):
        return SingleOperandHolder(NOT, self)

    def __eq__(self, other) -> bool:
        if not hasattr(self, "op1_class"):
            return self is other
        if hasattr(self, "op2_class"):
            return self.op1_class is other.op1_class and self.op2_class is other.op2_class
        else:
            return self.op1_class is other.op1_class


class SingleOperandHolder(OperationHolderMixin):
    def __init__(self, operator_class, op1_class):
        self.operator_class = operator_class
        self.op1_class = op1_class

    def __call__(self, *args, **kwargs):
        op1 = self.op1_class(*args, **kwargs)
        return self.operator_class(op1)


class OperandHolder(OperationHolderMixin):
    def __init__(self, operator_class, op1_class, op2_class):
        self.operator_class = operator_class
        self.op1_class = op1_class
        self.op2_class = op2_class

    def __call__(self, *args, **kwargs):
        op1 = self.op1_class(*args, **kwargs)
        op2 = self.op2_class(*args, **kwargs)
        return self.operator_class(op1, op2)


class AND:
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    def has_permission(self, request, resolver):
        return self.op1.has_permission(request, resolver) and self.op2.has_permission(
            request, resolver
        )

    def has_object_permission(self, request, resolver, obj):
        return self.op1.has_object_permission(
            request, resolver, obj
        ) and self.op2.has_object_permission(request, resolver, obj)


class OR:
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    def has_permission(self, request, resolver):
        return self.op1.has_permission(request, resolver) or self.op2.has_permission(
            request, resolver
        )

    def has_object_permission(self, request, resolver, obj):
        return self.op1.has_object_permission(
            request, resolver, obj
        ) or self.op2.has_object_permission(request, resolver, obj)


class NOT:
    def __init__(self, op1):
        self.op1 = op1

    def has_permission(self, request, resolver):
        return not self.op1.has_permission(request, resolver)

    def has_object_permission(self, request, resolver, obj):
        return not self.op1.has_object_permission(request, resolver, obj)


class BasePermissionMetaclass(OperationHolderMixin, type):
    pass


class BasePermission(metaclass=BasePermissionMetaclass):
    """
    A base class from which all permission classes should inherit.
    """

    def has_permission(self, request, resolver):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True

    def has_object_permission(self, request, resolver, obj):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True


class AllowAny(BasePermission):
    """
    Allow any access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(self, request, resolver):
        return True


class IsAuthenticated(BasePermission):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self, request, resolver):
        return bool(request.user and request.user.is_authenticated)


class IsAdminUser(BasePermission):
    """
    Allows access only to admin users.
    """

    def has_permission(self, request, resolver):
        return bool(request.user and request.user.is_staff)



from django.utils.translation import gettext_lazy as _


class ResolverException(Exception):
    pass


class NotFoundException(ResolverException):
    pass


class PermissionDenied(ResolverException):
    default_detail = _("You do not have permission to query this field.")
    default_code = "permission_denied"


class Throttled(ResolverException):
    pass
